import tortoise
from tortoise.expressions import Q

from .models import (
    Status, Type, Task, User, Comment, Category
)
from exceptions import common as common_exc


class BaseRepository:
    model: tortoise.Model

    async def get_list(self, **kwargs) -> list[tortoise.Model]:
        return await self.model.filter(**kwargs)

    async def get(self, id: int) -> tortoise.Model:
        try:
            return await self.model.get(id=id)

        except tortoise.exceptions.DoesNotExist as e:
            raise common_exc.NotFoundException(str(e))

    async def create(self, **kwargs) -> tortoise.Model:
        try:
            return await self.model.create(**kwargs)

        except tortoise.exceptions.IntegrityError as e:
            raise common_exc.CreateException(str(e))

    async def update(self, id: int, **kwargs) -> tortoise.Model:
        try:
            instance = await self.get(id=id)
            await instance.update_from_dict(kwargs).save()

            return instance

        except tortoise.exceptions.IntegrityError as e:
            raise common_exc.UpdateException(str(e))

    async def delete(self, id: int) -> None:
        try:
            instance = await self.get(id=id)
            await instance.delete()

        except tortoise.exceptions.IntegrityError as e:
            raise common_exc.DeleteException(str(e))


class StatusRepository(BaseRepository):
    model = Status


class TypeRepository(BaseRepository):
    model = Type


class TaskRepository(BaseRepository):
    model = Task

    async def get_list(
        self,
        title: str | None = None,
        status: str | None = None,
        category: str | None = None,
        priority: str | None = None
    ):
        query = self.model

        if title is not None:
            query = query.filter(title__icontains=title)

        if status is not None:
            query = query.filter(status__name__icontains=status)

        if category is not None:
            query = query.filter(category__name__icontains=category)

        if priority is not None:
            query = query.filter(priority=priority)

        return await query.all()

    async def search(self, keywords: str):
        query = self.model
        search_terms = keywords.split()

        conditions = [Q(title__icontains=term) | Q(description__icontains=term) for term in search_terms]

        combined_condition = conditions.pop()

        for condition in conditions:
            combined_condition |= condition

        query = query.filter(combined_condition)

        return await query.all()


class UserRepository(BaseRepository):
    model = User


class CommentRepository(BaseRepository):
    model = Comment


class CategoryRepository(BaseRepository):
    model = Category
